const std = @import("std");

const shell = @import("shell.zig");
const util = @import("util.zig");
const cmd = @import("cmd.zig");

const StdIO = shell.StdIO;
const Loop = shell.Loop;
const Shell = shell.Shell;

const PEDANTIC = false;

pub const BltnCtx = struct {
    loop: *Loop,
    cmd_name: []const u8,
    args: []const []const u8,
    allocator: std.mem.Allocator,
    shell: *Shell,
    stdio: StdIO,

    fn get_sopt(self: @This(), idx: usize) ?u8 {
        if (self.args.len < 1) return null;
        if (idx > self.args.len - 1) return null;
        const candidate = self.args[idx];
        if (candidate.len != 2) return null;
        if (candidate[0] != '-') return null;
        return candidate[1];
    }

    fn clone_arg(self: @This(), idx: usize) ![]const u8 {
        std.debug.assert(idx < self.args.len);
        return try self.allocator.dupe(u8, self.args[idx]);
    }

    fn clone_args(self: @This()) ![]const []const u8 {
        var al = std.ArrayList([]u8).init(self.allocator);
        for (self.args) |arg| try al.append(try self.allocator.dupe(u8, arg));
        return try al.toOwnedSlice();
    }

    fn usage_unknown(
        self: @This(),
        comptime B: type,
        arg: []const u8,
    ) error{UsageError} {
        self.stdio.err(B.USAGE_TEXT);
        self.stdio.err("\n\n");
        self.stdio.errf("unknown argument: '{s}'", .{arg});
        return error.UsageError;
    }

    fn usage_missing(
        self: @This(),
        comptime B: type,
        comptime label: []const u8,
    ) error{UsageError} {
        self.stdio.err(B.USAGE_TEXT);
        self.stdio.err("\n\n");
        self.stdio.err("no " ++ label ++ "?");
        return error.UsageError;
    }

    fn usage_toomany(
        self: @This(),
        comptime B: type,
    ) error{UsageError} {
        self.stdio.err(B.USAGE_TEXT);
        self.stdio.err("\n\n");
        self.stdio.err("too many arguments");
        return error.UsageError;
    }

    fn usage(
        self: @This(),
        comptime B: type,
        comptime fmt: []const u8,
        args: anytype,
    ) error{UsageError} {
        self.stdio.err(B.USAGE_TEXT);
        self.stdio.err("\n\n");
        self.stdio.errf(fmt, args);
        return error.UsageError;
    }
};

const ExitBltn = struct {
    const Self = @This();
    const Params = struct {
        es: u7,
    };

    const USAGE_TEXT =
        \\usage: exit [N]
    ;

    fn _parse(ctx: BltnCtx) !Params {
        if (ctx.args.len == 0) return Params{ .es = shell.EXIT_OK };
        if (ctx.args.len > 1) return ctx.usage_toomany(Self);
        const es_str = ctx.args[0];
        const es = std.fmt.parseInt(u7, es_str, 10) catch {
            return ctx.usage(
                Self,
                "N must be a base-10 integer between 0 and 127: '{s}'",
                .{es_str},
            );
        };
        return Params{ .es = es };
    }

    fn run(ctx: BltnCtx) !void {
        const params: Params = try Self._parse(ctx);
        ctx.shell.exit(params.es);
    }
};

const EchoBltn = struct {
    const Self = @This();
    const Params = struct {
        operands: []const []const u8,
    };

    const USAGE_TEXT =
        \\usage: echo [STRING..]
    ;

    fn _parse(ctx: BltnCtx) !Params {
        return Params{ .operands = try ctx.clone_args() };
    }

    fn run(ctx: BltnCtx) !void {
        const params: Params = try Self._parse(ctx);
        const w = ctx.stdio.stdout_f.writer().any();
        for (params.operands, 0..) |operand, idx| {
            if (idx > 0) try w.writeByte(' ');
            try w.writeAll(operand);
        }
        try w.writeByte('\n');
    }
};

const TypeBltn = struct {
    const Self = @This();
    const Params = struct {
        cmd_name: []const u8,
    };

    const USAGE_TEXT =
        \\usage: type COMMAND_NAME
    ;

    fn _parse(ctx: BltnCtx) !Params {
        if (ctx.args.len == 0) return ctx.usage_missing(Self, "COMMAND_NAME");
        if (ctx.args.len > 1) return ctx.usage_toomany(Self);
        const cmd_name = try ctx.clone_arg(0);
        if (std.mem.indexOfAny(u8, cmd_name, shell.CMD_NAME_INVALID)) |pos| return ctx.usage(
            Self,
            "COMMAND_NAME contains invalid character: {x} at position {d}",
            .{ cmd_name[pos], pos },
        );
        return Params{ .cmd_name = cmd_name };
    }

    fn run(ctx: BltnCtx) !void {
        const params = try Self._parse(ctx);
        if (Bltn.exists(params.cmd_name)) {
            try ctx.stdio.outf("{s} is a shell builtin", .{params.cmd_name});
        } else if (try cmd.find_exec_path(params.cmd_name, ctx.allocator)) |exec_path| {
            try ctx.stdio.outf("{s} is {s}", .{ params.cmd_name, exec_path });
        } else {
            try ctx.stdio.outf("{s}: not found", .{params.cmd_name});
        }
    }
};

const PwdBltn = struct {
    const Self = @This();
    const Params = struct {
        mode: Mode,
    };
    const Mode = enum { L, P };
    const USAGE_TEXT =
        \\usage: pwd [-L|-P]
    ;

    fn _parse(ctx: BltnCtx) !Params {
        var cur: usize = 0;
        var mode: Mode = .L;
        while (cur < ctx.args.len) {
            const arg = ctx.args[cur];
            const maybe_opt_letter = ctx.get_sopt(cur);
            if (maybe_opt_letter == null) return ctx.usage_unknown(Self, arg);
            switch (maybe_opt_letter.?) {
                'L' => mode = .L,
                'P' => mode = .P,
                else => return ctx.usage_unknown(Self, arg),
            }
            cur += 1;
        }
        return Params{ .mode = mode };
    }

    fn run(ctx: BltnCtx) !void {
        const params = try Self._parse(ctx);
        const out = b: {
            if (PEDANTIC) {
                const PwdComputer = @import("pedantic.zig").PwdComputer;
                const pc = try PwdComputer.create(ctx.allocator);
                defer pc.destroy();
                switch (params.mode) {
                    .P => break :b pc.getpath_P(),
                    .L => break :b pc.getpath_L(),
                }
            } else {
                break :b try std.process.getCwdAlloc(ctx.allocator);
            }
        };
        try ctx.stdio.out(out);
    }
};

const CdBltn = struct {
    const USAGE_TEXT =
        \\usage: cd PATH
    ;
    const Self = @This();
    const Target = union(enum) {
        path: []const u8,
        home: void,
        old: void,
    };

    fn _parse_target(ctx: BltnCtx) !Target {
        if (ctx.args.len == 0) return Target{ .home = {} };
        if (ctx.args.len > 1) return ctx.usage_toomany(Self);
        if (ctx.args[0].len == 1 and ctx.args[0][0] == '-') return Target{ .old = {} };
        return Target{ .path = try ctx.clone_arg(0) };
    }

    fn _err2msg(err: anyerror) ?[]const u8 {
        return (switch (err) {
            error.AccessDenied => "Access denied",
            error.FileNotFound => "No such file or directory",
            error.NotDir => "Not a directory",
            error.BadPathName => "Invalid path name",
            else => null,
        });
    }

    fn _chdir(ctx: BltnCtx, path: []const u8) !void {
        const initial_path = try std.process.getCwdAlloc(ctx.allocator);
        defer ctx.allocator.free(initial_path);
        const untilded = try util.realpathWithTilde(ctx.allocator, path);
        try std.process.changeCurDir(untilded);
        try ctx.shell.setOldPwd(initial_path);
    }

    fn run(ctx: BltnCtx) !void {
        const target = try Self._parse_target(ctx);
        const path = switch (target) {
            .path => |path| path,
            .home => try std.process.getEnvVarOwned(ctx.allocator, "HOME"),
            .old => ctx.shell.vars.OLDPWD orelse {
                ctx.stdio.err("cd: OLDPWD not set");
                return;
            },
        };
        Self._chdir(ctx, path) catch |e| {
            const msg = Self._err2msg(e) orelse return e;
            ctx.stdio.errf("cd: {s}: {s}", .{ path, msg });
        };
    }
};

const _VarsBltn = struct {
    const USAGE_TEXT =
        \\usage: _vars
    ;
    const Self = @This();
    const Params = struct {};

    fn run(ctx: BltnCtx) !void {
        if (ctx.args.len > 0) return ctx.usage_toomany(Self);
        if (ctx.shell.vars.OLDPWD) |path| ctx.stdio.errf("OLDPWD={s}", .{path});
        ctx.stdio.errf("ES={d}", .{ctx.shell.vars.ES});
    }
};

const Bltn = enum {
    exit,
    echo,
    type,
    pwd,
    cd,
    _vars,

    fn exists(cmd_name: []const u8) bool {
        if (Bltn.select(cmd_name)) |_| return true;
        return false;
    }

    fn select(cmd_name: []const u8) ?Bltn {
        if (std.mem.eql(u8, cmd_name, "exit")) return .exit;
        if (std.mem.eql(u8, cmd_name, "echo")) return .echo;
        if (std.mem.eql(u8, cmd_name, "type")) return .type;
        if (std.mem.eql(u8, cmd_name, "pwd")) return .pwd;
        if (std.mem.eql(u8, cmd_name, "cd")) return .cd;
        if (std.mem.eql(u8, cmd_name, "_vars")) return ._vars;
        return null;
    }
};

pub fn run_bltn(ctx: BltnCtx) !void {
    const bltn = Bltn.select(ctx.cmd_name) orelse return error.NoSuchBltn;
    std.log.debug("running bltn: {any}", .{bltn});
    return switch (bltn) {
        .exit => ExitBltn.run(ctx),
        .echo => EchoBltn.run(ctx),
        .type => TypeBltn.run(ctx),
        .pwd => PwdBltn.run(ctx),
        .cd => CdBltn.run(ctx),
        ._vars => _VarsBltn.run(ctx),
    };
}
