const std = @import("std");

const bltn_ = @import("bltn.zig");
const cmd_ = @import("cmd.zig");
const syntax = @import("syntax.zig");
const util = @import("util.zig");

const Bltn = bltn_.Bltn;
const RedirOp = syntax.RedirOp;
const run_command = cmd_.run_command;

pub const EXIT_OK = 0;
pub const EXIT_NO = 1;
pub const EXIT_USAGE_ERROR = 2;
pub const EXIT_ERROR = 3;
pub const EXIT_ERRPANIC = 4;
pub const EXIT_IOPANIC = 5;

pub const CMD_NAME_INVALID = "/\x00\n";

pub const DEVEL_MODE: enum { devel, release } = .devel;
pub const DEVEL_MODE_SHOW_BUFFER: bool = false;

const BUFSIZE_REPL = 8192;
const BUFSIZE_INPUT = 1024;

pub const StdIO = struct {
    stdin_f: std.fs.File,
    stdout_f: std.fs.File,
    stderr_f: std.fs.File,
    shell_stderr_f: std.fs.File,
    stdin_redir: ?RedirOp,
    stdout_redir: ?RedirOp,
    stderr_redir: ?RedirOp,

    fn _handle_stderr_error(self: @This(), e: anyerror) void {
        // in case of external stdin, we can fall
        if (self.stderr_redir == null) iopanic(); // not redirected
        const path = self.stderr_redir.?.path;
        const w = self.shell_stderr_f.writer();
        w.print("cannot write to redirected stderr: {!} when writing to {s}\n", .{ e, path }) catch iopanic();
    }

    /// Try to write details of error *e* to file handle *err_f* and die.
    /// This exits the process with EXIT_ERRPANIC, except when the write fails
    /// in which case it exits with EXIT_IOPANIC.
    fn errpanic(err_f: std.fs.File, e: anyerror) noreturn {
        err_f.writeAll("panic: ") catch iopanic();
        err_f.writeAll(@errorName(e)) catch iopanic();
        err_f.writeAll("\n") catch iopanic();
        std.process.exit(EXIT_ERRPANIC);
    }

    pub const Files = struct {
        stdin: ?std.fs.File = null,
        stdout: ?std.fs.File = null,
        stderr: ?std.fs.File = null,
    };

    fn redird_files(self: @This()) Files {
        var fs = Files{};
        if (self.stdin_redir != null) fs.stdin = self.stdin_f;
        if (self.stdout_redir != null) fs.stdout = self.stdout_f;
        if (self.stderr_redir != null) fs.stderr = self.stderr_f;
        return fs;
    }

    fn close(self: @This()) void {
        if (self.stdin_redir) |op| {
            std.log.debug("closing stdin redirect: {s}", .{op.path});
            self.stdin_f.close();
        }
        if (self.stdout_redir) |op| {
            std.log.debug("closing stdout redirect: {s}", .{op.path});
            self.stdout_f.close();
        }
        if (self.stderr_redir) |op| {
            std.log.debug("closing stderr redirect: {s}", .{op.path});
            self.stderr_f.close();
        }
    }

    pub fn err(self: @This(), msg: []const u8) void {
        self.stderr_f.writeAll(msg) catch |e| self._handle_stderr_error(e);
        self.stderr_f.writeAll("\n") catch |e| self._handle_stderr_error(e);
    }

    pub fn errf(self: @This(), comptime fmt: []const u8, args: anytype) void {
        self.stderr_f.writer().print(fmt, args) catch |e| self._handle_stderr_error(e);
        self.stderr_f.writeAll("\n") catch |e| self._handle_stderr_error(e);
    }

    pub fn outw(self: @This(), msg: []const u8) !void {
        try self.stdout_f.writeAll(msg);
    }

    pub fn out(self: @This(), msg: []const u8) !void {
        try self.stdout_f.writeAll(msg);
        try self.stdout_f.writeAll("\n");
    }

    pub fn outf(self: @This(), comptime fmt: []const u8, args: anytype) !void {
        try self.stdout_f.writer().print(fmt, args);
        try self.stdout_f.writeAll("\n");
    }
};

/// Context of single REPL iteration
pub const Loop = struct {
    shell: *Shell,
    allocator: std.mem.Allocator,

    /// Get input from user
    fn _get_input(self: *@This()) ![]const u8 {
        try self.shell.outw("$ ");
        const r = self.shell.stdin_f.reader();
        return try r.readUntilDelimiter(self.shell.input_buffer.?, '\n');
    }

    /// Open path for IO redirect
    fn _open_redir_file(op: RedirOp) !std.fs.File {
        const cwd_fd = std.fs.cwd();
        switch (op.dir) {
            .in => return try cwd_fd.openFile(op.path, .{ .mode = .read_only }),
            .out => return try cwd_fd.createFile(op.path, .{}),
            .app => {
                // zls fails here; thinks `res` is `File` correct is `?File`
                const res: ?std.fs.File = cwd_fd.openFile(op.path, .{ .mode = .write_only }) catch |e| switch (e) {
                    error.FileNotFound => null,
                    else => return e,
                };
                if (res) |fd| {
                    try fd.seekFromEnd(0);
                    return fd;
                }
                return try cwd_fd.createFile(op.path, .{});
            },
        }
    }

    /// Create StdIO object with redirects applied and the proper filehandles open
    fn _create_stdio(
        self: @This(),
        stdin_redir: ?RedirOp,
        stdout_redir: ?RedirOp,
        stderr_redir: ?RedirOp,
    ) !StdIO {
        return StdIO{
            .stdin_f = if (stdin_redir) |op| try _open_redir_file(op) else self.shell.stdin_f,
            .stdout_f = if (stdout_redir) |op| try _open_redir_file(op) else self.shell.stdout_f,
            .stderr_f = if (stderr_redir) |op| try _open_redir_file(op) else self.shell.stderr_f,
            .shell_stderr_f = self.shell.stderr_f,
            .stdin_redir = stdin_redir,
            .stdout_redir = stdout_redir,
            .stderr_redir = stderr_redir,
        };
    }

    /// Do one loop of REPL
    fn run(self: *@This()) !void {
        const user_input = self._get_input() catch |e| switch (e) {
            error.EndOfStream => self.shell.exit_eof(),
            else => return e,
        };

        if (user_input.len == 0) return;
        const stmt = try syntax.parse_stmt(user_input, self.allocator);
        if (stmt.argv.len == 0) return;

        const stdio = try self._create_stdio(
            stmt.stdin_redir,
            stmt.stdout_redir,
            stmt.stderr_redir,
        );
        defer stdio.close();

        const ctx = bltn_.BltnCtx{
            .shell = self.shell,
            .stdio = stdio,
            .loop = self,
            .cmd_name = stmt.argv[0],
            .args = stmt.argv[1..],
            .allocator = self.allocator,
        };

        bltn_.run_bltn(ctx) catch |e| switch (e) {
            error.UsageError => {},
            error.NoSuchBltn => {
                const es = run_command(
                    stmt.argv,
                    stdio.redird_files(),
                    self.allocator,
                ) catch |run_e| switch (run_e) {
                    error.NoSuchCommand => {
                        self.shell.errf("{s}: command not found", .{stmt.argv[0]});
                        return;
                    },
                    else => return run_e,
                };
                _ = es;
            },
            else => return e,
        };
    }
};

/// Just die after realizing even IO does not work
///
/// Use this as last resort, when catching from a call which is not expected
/// to have any other mode of failure than stderr writing error.
///
/// In any other cases (eg. when the call involves formatting, ie. the failure
/// could be something like memory allocation), catch the error and use
/// `errpanic(fh, e)` at least, so that user has a chance of understanding what
/// is wrong.
///
/// Since EXIT_IOPANIC is not used anywhere else, at least seeing that exit
/// status can hint towards the fact that there was error but stderr was broken
/// so we could not provide any info.
pub fn iopanic() noreturn {
    std.process.exit(EXIT_IOPANIC);
}

/// Try to write details of error *e* to file handle *err_f* and die.
/// This exits the process with EXIT_ERRPANIC, except when the write fails
/// in which case it exits with EXIT_IOPANIC.
pub fn errpanic(err_f: std.fs.File, e: anyerror) noreturn {
    err_f.writeAll("panic: ") catch iopanic();
    err_f.writeAll(@errorName(e)) catch iopanic();
    err_f.writeAll("\n") catch iopanic();
    std.process.exit(EXIT_ERRPANIC);
}

pub const Shell = struct {
    stdin_f: std.fs.File,
    stdout_f: std.fs.File,
    stderr_f: std.fs.File,
    allocator: std.mem.Allocator,
    repl_buffer: ?[]u8 = null,
    input_buffer: ?[]u8 = null,
    vars: Vars,

    const Vars = struct {
        OLDPWD: ?[]const u8 = null,
        ES: u8 = 0,
    };

    pub fn init(allocator: std.mem.Allocator) Shell {
        return Shell{
            .stdout_f = std.io.getStdOut(),
            .stdin_f = std.io.getStdIn(),
            .stderr_f = std.io.getStdErr(),
            .allocator = allocator,
            .vars = Vars{},
        };
    }

    pub fn setOldPwd(self: *@This(), path: []const u8) !void {
        if (self.vars.OLDPWD) |pwd| {
            self.allocator.free(pwd);
        }
        self.vars.OLDPWD = try self.allocator.dupe(u8, path);
    }

    fn errf(self: @This(), comptime fmt: []const u8, args: anytype) void {
        self.stderr_f.writer().print(fmt, args) catch |e| errpanic(self.stderr_f, e);
        self.stderr_f.writeAll("\n") catch iopanic();
    }

    fn outw(self: @This(), msg: []const u8) !void {
        try self.stdout_f.writeAll(msg);
    }

    fn exit_eof(self: @This()) noreturn {
        self.stdout_f.writeAll("\n") catch iopanic();
        std.process.exit(EXIT_OK);
    }

    pub fn exit(self: @This(), es: u7) noreturn {
        _ = self;
        std.process.exit(@as(u8, es));
    }

    fn repl(self: *@This()) !void {
        var fba = std.heap.FixedBufferAllocator.init(self.repl_buffer.?);
        while (true) {
            fba.reset();
            var loop = Loop{
                .shell = self,
                .allocator = fba.allocator(),
            };
            try loop.run();
        }
    }

    pub fn run(self: *@This()) !void {
        self.repl_buffer = self.allocator.alloc(u8, BUFSIZE_REPL) catch |e| errpanic(self.stderr_f, e);
        self.input_buffer = self.allocator.alloc(u8, BUFSIZE_INPUT) catch |e| errpanic(self.stderr_f, e);
        defer self.allocator.free(self.repl_buffer.?);
        defer self.allocator.free(self.input_buffer.?);
        _ = self.repl() catch |e| {
            switch (DEVEL_MODE) {
                .devel => {
                    if (DEVEL_MODE_SHOW_BUFFER) self.errf("REPL buffer: {s}", .{util.qb(self.repl_buffer.?, 100000)});
                    return e;
                },
                .release => errpanic(self.stderr_f, e),
            }
        };
    }
};
