const std = @import("std");

const util = @import("util.zig");

const DEBUG = false;

pub const RedirOp = struct {
    path: []const u8,
    dir: Dir,
    const Dir = enum { in, out, app };
};

const Statement = struct {
    argv: []const []const u8,
    stdin_redir: ?RedirOp = null,
    stdout_redir: ?RedirOp = null,
    stderr_redir: ?RedirOp = null,

    fn _check_path(op: ?RedirOp) !void {
        if (op == null) return;
        if (op.?.path.len == 0) return error.InvalidRedirectionPath; // empty path?
    }

    fn init(
        argv: []const []const u8,
        stdin_redir: ?RedirOp,
        stdout_redir: ?RedirOp,
        stderr_redir: ?RedirOp,
    ) !Statement {
        try Statement._check_path(stdin_redir);
        try Statement._check_path(stdout_redir);
        try Statement._check_path(stderr_redir);
        if (argv.len == 0) {
            if (stdin_redir != null) return error.RedirectWithoutCommand;
            if (stdout_redir != null) return error.RedirectWithoutCommand;
            if (stderr_redir != null) return error.RedirectWithoutCommand;
        }
        return Statement{
            .argv = argv,
            .stdin_redir = stdin_redir,
            .stdout_redir = stdout_redir,
            .stderr_redir = stderr_redir,
        };
    }
};

const TokenTag = enum {
    gap,
    fd,
    lt,
    gt,
    str,
};

const Token = struct {
    tag: TokenTag,
    bgn: usize,
    len: usize,
    fn init(tag: TokenTag, bgn: usize, len: usize) Token {
        return Token{
            .tag = tag,
            .bgn = bgn,
            .len = len,
        };
    }
    fn get(self: @This(), src: []const u8) ![]const u8 {
        const bgn = self.bgn;
        const end = self.bgn + self.len;
        if (bgn >= src.len) return error.TokenOverflow;
        if (end > src.len) return error.TokenOverflow;
        return src[bgn..end];
    }
};

fn show_token_minimap(
    src: []const u8,
    tokens: []Token,
) void {
    var al = std.ArrayList(u8).init(std.heap.page_allocator);
    defer al.deinit();
    std.log.warn("src=|{s}|", .{src});
    for (tokens) |t| {
        const w = al.writer().any();
        w.writeByteNTimes(' ', t.bgn) catch {};
        w.writeByteNTimes('^', t.len) catch {};
        w.writeByteNTimes(' ', src.len - t.len - t.bgn) catch {};
        std.log.warn("     {s} .. .{s}@{d}+{d}", .{ al.items, @tagName(t.tag), t.bgn, t.len });
        al.resize(0) catch {};
    }
}

const TokenIterator = struct {
    allocator: std.mem.Allocator,
    bitr: ByteItr,
    queue: TokenQueue,

    const TokenQueue = std.fifo.LinearFifo(Token, .Dynamic);

    const CharClass = enum {
        sp,
        do,
        fd,
        lt,
        gt,
        bs,
        dq,
        sq,
        ch,
        fn select(c: u8) CharClass {
            return switch (c) {
                '$' => .do,
                ' ', '\t' => .sp,
                '0'...'9' => .fd,
                '>' => .gt,
                '<' => .lt,
                '\\' => .bs,
                '"' => .dq,
                '\'' => .sq,
                else => .ch,
            };
        }
    };

    const ByteItr = struct {
        buff: []const u8,
        cur: usize = 0,

        fn next(self: *@This()) ?u8 {
            if (self.cur >= self.buff.len) return null;
            const out = self.buff[self.cur];
            self.cur += 1;
            return out;
        }

        fn peek0(self: *@This()) ?u8 {
            if (self.cur >= self.buff.len) return null;
            return self.buff[self.cur];
        }

        fn jump_until_none_or_end(self: *@This(), chars: []const u8) usize {
            const offset = std.mem.indexOfNone(
                u8,
                self.buff[self.cur..],
                chars,
            ) orelse self.buff[self.cur..].len;
            self.cur += offset;
            return offset;
        }

        fn jump_until(self: *@This(), chars: []const u8) ?usize {
            const offset = std.mem.indexOfAny(
                u8,
                self.buff[self.cur..],
                chars,
            ) orelse return null;
            self.cur += offset;
            return offset;
        }
    };

    fn init(buff: []const u8, allocator: std.mem.Allocator) TokenIterator {
        return TokenIterator{
            .allocator = allocator,
            .bitr = ByteItr{ .cur = 0, .buff = buff },
            .queue = TokenQueue.init(allocator),
        };
    }

    fn _send_gap(self: *@This()) !void {
        const bgn = self.bitr.cur;
        const len = self.bitr.jump_until_none_or_end(" \t");
        try self._emit_token(.gap, bgn, len);
    }

    fn _send_do(self: *@This()) !void {
        const c = self.bitr.next() orelse return error.InternalError;
        if (c != '$') return error.InternalError;
        try self._emit_token(.str, self.bitr.cur - 1, 1);
    }

    fn _emit_token(self: *@This(), tag: TokenTag, bgn: usize, len: usize) !void {
        const t = Token{ .tag = tag, .bgn = bgn, .len = len };
        std.log.debug(" >>>> buff=|{s}|", .{self.bitr.buff});
        std.log.debug(" >>>> Token{{ .tag=.{s} .bgn={d} .len={d} }}", .{ @tagName(t.tag), t.bgn, t.len });
        try self.queue.writeItem(t);
    }

    fn _send_fd(self: *@This()) !void {
        const c = self.bitr.next() orelse return error.InternalError;
        if (CharClass.select(c) != .fd) return error.InternalError;
        try self._emit_token(.fd, self.bitr.cur - 1, 1);
    }

    fn _send_gt(self: *@This()) !void {
        const c = self.bitr.next() orelse return error.InternalError;
        if (c != '>') return error.InternalError;
        try self._emit_token(.gt, self.bitr.cur - 1, 1);
    }

    fn _send_lt(self: *@This()) !void {
        const c = self.bitr.next() orelse return error.InternalError;
        if (c != '<') return error.InternalError;
        try self._emit_token(.lt, self.bitr.cur - 1, 1);
    }

    fn _send_escaped(self: *@This()) !void {
        std.log.debug("TokenIterator._send_escaped():.rest=|{s}|", .{self.bitr.buff[self.bitr.cur..]});
        const c = self.bitr.next() orelse return error.InternalError;
        std.log.debug("TokenIterator._send_escaped():|rest=|{s}|", .{self.bitr.buff[self.bitr.cur..]});
        if (c != '\\') return error.InternalError;
        _ = self.bitr.next() orelse return error.MissingEscapedChar;
        try self._emit_token(.str, self.bitr.cur - 1, 1);
    }

    fn _send_single_quote(self: *@This()) !void {
        std.log.debug("TokenIterator._send_single_quote():.rest=|{s}|", .{self.bitr.buff[self.bitr.cur..]});
        const left_idx = self.bitr.cur;
        const c = self.bitr.next() orelse return error.InternalError;
        std.log.debug("TokenIterator._send_single_quote():|rest=|{s}|", .{self.bitr.buff[self.bitr.cur..]});
        if (c != '\'') return error.InternalError;
        const len = self.bitr.jump_until("'") orelse return error.MissingEndSingleQuote;
        try self._emit_token(.str, left_idx + 1, len);
        _ = self.bitr.next() orelse return error.InternalError;
        std.log.debug("TokenIterator._send_single_quote():'rest=|{s}|", .{self.bitr.buff[self.bitr.cur..]});
    }

    fn _send_double_quote(self: *@This()) !void {
        const Span = struct { bgn: usize, len: usize = 0 };
        const c = self.bitr.next() orelse return error.InternalError;
        if (c != '"') return error.InternalError;
        std.log.debug("TokenIterator._send_double_quote(): .. buff=|{s}|", .{self.bitr.buff});
        var span = Span{ .bgn = self.bitr.cur };
        std.log.debug("TokenIterator._send_double_quote(): .. span=S{{{d} {d}}}", .{ span.bgn, span.len });
        // eg. `"aaaaa\\bbbb\$cccc\"dddd"` will go as 4 spans
        while (true) {
            span.len += self.bitr.jump_until("\"\\") orelse return error.MissingEndDoubleQuote;
            std.log.debug("TokenIterator._send_double_quote(): :: span=S{{{d} {d}}}", .{ span.bgn, span.len });
            const char = self.bitr.next() orelse return error.MissingEndDoubleQuote;
            std.log.debug("TokenIterator._send_double_quote(): :: char={c}", .{char});
            switch (char) {
                '"' => {
                    std.log.debug("END", .{});
                    try self._emit_token(.str, span.bgn, span.len);
                    return;
                },
                '\\' => {
                    std.log.debug("BS", .{});
                    const char1 = self.bitr.peek0() orelse return error.MissingEndDoubleQuote;
                    std.log.debug("TokenIterator._send_double_quote(): :: char1={c}", .{char1});
                    switch (char1) {
                        '"', '\\', '$' => {
                            std.log.debug("BS>POWERFUL", .{});
                            // `\` has special meaning: forget it and start new span with char1
                            if (span.len > 0) try self._emit_token(.str, span.bgn, span.len);
                            span = Span{ .bgn = self.bitr.cur, .len = 1 };
                            std.log.debug("TokenIterator._send_double_quote(): :: span=S{{{d} {d}}}", .{ span.bgn, span.len });
                            _ = self.bitr.next();
                        },
                        else => {
                            // `\` treated as normal character; just continue the span
                            std.log.debug("BS>PLAIN", .{});
                            span.len += 1;
                        },
                    }
                },
                else => return error.InternalError, // impossible since we jumped until `"` or `\`
            }
        }
    }

    fn _send_literals(self: *@This()) !void {
        // !! initial double quote must already have been removed !!
        var itr = ByteItr{ .buff = self.bitr.buff[self.bitr.cur..] };
        var len: usize = 0;
        const bgn = self.bitr.cur;
        while (itr.next()) |c| {
            const cls = CharClass.select(c);
            if (cls == .ch) len += 1 else break;
        }
        if (len == 0) return error.InternalError;
        try self._emit_token(.str, bgn, len);
        self.bitr.cur += len;
    }

    fn _maybe_send_next(self: *@This()) !void {
        const c = self.bitr.peek0() orelse return;
        const cls = CharClass.select(c);
        std.log.debug("TokenIterator.next():c={c} cls=.{s}", .{ c, @tagName(cls) });
        switch (cls) {
            .sp => try self._send_gap(),
            .fd => try self._send_fd(),
            .do => try self._send_do(),
            .gt => try self._send_gt(),
            .lt => try self._send_lt(),
            .bs => try self._send_escaped(),
            .sq => try self._send_single_quote(),
            .dq => try self._send_double_quote(),
            .ch => try self._send_literals(),
        }
    }

    fn peek(self: *@This()) !?Token {
        if (self.queue.readableLength() < 1) try self._maybe_send_next();
        if (self.queue.readableLength() < 1) return null;
        return self.queue.readableSlice(0)[0];
    }

    fn peek2_is(self: *@This(), tag0: TokenTag, tag1: TokenTag) !bool {
        if (self.queue.readableLength() < 2) try self._maybe_send_next();
        if (self.queue.readableLength() < 2) try self._maybe_send_next();
        if (self.queue.readableLength() < 2) return false;
        const win = self.queue.readableSlice(0)[0..2];
        return win[0].tag == tag0 and win[1].tag == tag1;
    }

    fn peek1_is(self: *@This(), tag: TokenTag) !bool {
        if (self.queue.readableLength() < 1) try self._maybe_send_next();
        if (self.queue.readableLength() < 1) return false;
        return self.queue.readableSlice(0)[0].tag == tag;
    }

    fn next(self: *@This()) !?Token {
        if (self.queue.readItem()) |token| return token;
        try self._maybe_send_next();
        return self.queue.readItem();
    }

    fn all(self: *@This()) ![]Token {
        var al = std.ArrayList(Token).init(self.allocator);
        while (try self.next()) |t| try al.append(t);
        const tokens = try al.toOwnedSlice();
        // show_token_minimap(self.bitr.buff, tokens);
        return tokens;
    }
};

pub fn tokenize(buff: []const u8, allocator: std.mem.Allocator) ![]Token {
    var titr = TokenIterator.init(buff, allocator);
    return try titr.all();
}

fn show_extra_tokens(comptime label: []const u8, buff: []const u8, tokens: []const Token) void {
    std.log.err(label ++ "buff: buff=|{s}|", .{buff});
    for (tokens) |t| std.log.err(label ++ "tag=.{s} bgn: {d} len:{d}", .{ @tagName(t.tag), t.bgn, t.len });
}

fn fmt_token(t: Token) []const u8 {
    return std.fmt.allocPrint(
        std.heap.page_allocator,
        ".{s}@{d}+{d}",
        .{ @tagName(t.tag), t.bgn, t.len },
    ) catch return "{FORMATTING_ERROR}";
}

fn tokenize_t(buff: []const u8, oracle: []const Token) !void {
    const result = try tokenize(buff, std.heap.page_allocator);
    errdefer show_token_minimap(buff, result);
    try util.assert_equal_slices(
        Token,
        &fmt_token,
        oracle,
        result,
    );
}

fn tokenize_et(buff: []const u8, oracle: anyerror) !void {
    const result = tokenize(buff, std.heap.page_allocator);
    try std.testing.expectEqual(oracle, result);
}

test "tokenize: epsilon" {
    try tokenize_t("", &.{});
}

test "tokenize: whitespace" {
    try tokenize_t(" ", &.{Token.init(.gap, 0, 1)});
    try tokenize_t("\t", &.{Token.init(.gap, 0, 1)});
    try tokenize_t("  ", &.{Token.init(.gap, 0, 2)});
    try tokenize_t("\t ", &.{Token.init(.gap, 0, 2)});
    try tokenize_t(" \t", &.{Token.init(.gap, 0, 2)});
    try tokenize_t("\t\t", &.{Token.init(.gap, 0, 2)});
    try tokenize_t(" \t ", &.{Token.init(.gap, 0, 3)});
    try tokenize_t("\t \t", &.{Token.init(.gap, 0, 3)});
}

test "tokenize: mixed" {
    try tokenize_t("foo  ", &.{
        Token.init(.str, 0, 3),
        Token.init(.gap, 3, 2),
    });
    try tokenize_t("  foo  ", &.{
        Token.init(.gap, 0, 2),
        Token.init(.str, 2, 3),
        Token.init(.gap, 5, 2),
    });
    try tokenize_t("  ''  ", &.{
        Token.init(.gap, 0, 2),
        Token.init(.str, 3, 0),
        Token.init(.gap, 4, 2),
    });
    try tokenize_t("  ''''  ", &.{
        Token.init(.gap, 0, 2),
        Token.init(.str, 3, 0),
        Token.init(.str, 5, 0),
        Token.init(.gap, 6, 2),
    });
    try tokenize_t("  ''\t''  ", &.{
        Token.init(.gap, 0, 2),
        Token.init(.str, 3, 0),
        Token.init(.gap, 4, 1),
        Token.init(.str, 6, 0),
        Token.init(.gap, 7, 2),
    });
}

test "tokenize: single quotes" {
    try tokenize_et("'", error.MissingEndSingleQuote);
    try tokenize_et("'''", error.MissingEndSingleQuote); // quoting `'` (single quote) - not possible
    try tokenize_t("'A Z'", &.{Token.init(.str, 1, 3)}); // quoting `A Z` (A, space, Z)
    try tokenize_t("'A\tZ'", &.{Token.init(.str, 1, 3)}); // quoting 0x09 (A, tab, Z)
    try tokenize_t("'A`Z'", &.{Token.init(.str, 1, 3)}); // quoting 0x96 (A, backtick, Z)
    try tokenize_t("'A\\Z'", &.{Token.init(.str, 1, 3)}); // quoting `A\Z` (A, backslash, Z)
    try tokenize_t("'A\"Z'", &.{Token.init(.str, 1, 3)}); // quoting `A"Z` (A, double quote, Z)
    try tokenize_t("'AMZ'", &.{Token.init(.str, 1, 3)}); // quoting `AMZ` (normal string)
}

test "tokenize: backslash escaping" {
    try tokenize_t("\\!", &.{Token.init(.str, 1, 1)}); // backslash escaping `!` (exclamation mark)
    try tokenize_t("\\$", &.{Token.init(.str, 1, 1)}); // backslash escaping `$` (dollar sign)
    try tokenize_t("\\&", &.{Token.init(.str, 1, 1)}); // backslash escaping `&` (ampersand)
    try tokenize_t("\\(", &.{Token.init(.str, 1, 1)}); // backslash escaping `(` (opening bracket)
    try tokenize_t("\\)", &.{Token.init(.str, 1, 1)}); // backslash escaping `)` (closing bracket)
    try tokenize_t("\\*", &.{Token.init(.str, 1, 1)}); // backslash escaping `*` (asterisk)
    try tokenize_t("\\;", &.{Token.init(.str, 1, 1)}); // backslash escaping `;` (semicolon)
    try tokenize_t("\\?", &.{Token.init(.str, 1, 1)}); // backslash escaping `?` (question mark)
    try tokenize_t("\\`", &.{Token.init(.str, 1, 1)}); // backslash escaping ``` (tick mark)
    try tokenize_t("\\{", &.{Token.init(.str, 1, 1)}); // backslash escaping `{` (opening curly brace)
    try tokenize_t("\\>", &.{Token.init(.str, 1, 1)}); // backslash escaping `>` (greater than)
    try tokenize_t("\\<", &.{Token.init(.str, 1, 1)}); // backslash escaping `<` (less than)
    try tokenize_t("\\|", &.{Token.init(.str, 1, 1)}); // backslash escaping `|` (pipe symbol)
    try tokenize_t("\\}", &.{Token.init(.str, 1, 1)}); // backslash escaping `}` (closing curly brace)
    try tokenize_t("\\ ", &.{Token.init(.str, 1, 1)}); // backslash escaping ` ` (space)
    try tokenize_t("\\'", &.{Token.init(.str, 1, 1)}); // backslash escaping `'` (single quote)
    try tokenize_t("\\\"", &.{Token.init(.str, 1, 1)}); // backslash escaping `"` (double quote)
    try tokenize_t("\\\\", &.{Token.init(.str, 1, 1)}); // backslash escaping `\` (backslash)
    try tokenize_t("\\\t", &.{Token.init(.str, 1, 1)}); // backslash escaping `\t` (tab)
    try tokenize_t("\\\n", &.{Token.init(.str, 1, 1)}); // backslash escaping `\n` (newline)
    try tokenize_t("\\A", &.{Token.init(.str, 1, 1)}); // backslash escaping `A` (letter A)
}

test "tokenize_t: single quoting" {
    try tokenize_t("'!'", &.{Token.init(.str, 1, 1)}); // single quoting `!` (exclamation mark)
    try tokenize_t("'$'", &.{Token.init(.str, 1, 1)}); // single quoting `$` (dollar sign)
    try tokenize_t("'&'", &.{Token.init(.str, 1, 1)}); // single quoting `&` (ampersand)
    try tokenize_t("'('", &.{Token.init(.str, 1, 1)}); // single quoting `(` (opening bracket)
    try tokenize_t("')'", &.{Token.init(.str, 1, 1)}); // single quoting `)` (closing bracket)
    try tokenize_t("'*'", &.{Token.init(.str, 1, 1)}); // single quoting `*` (asterisk)
    try tokenize_t("';'", &.{Token.init(.str, 1, 1)}); // single quoting `;` (semicolon)
    try tokenize_t("'?'", &.{Token.init(.str, 1, 1)}); // single quoting `?` (question mark)
    try tokenize_t("'`'", &.{Token.init(.str, 1, 1)}); // single quoting ``` (tick mark)
    try tokenize_t("'{'", &.{Token.init(.str, 1, 1)}); // single quoting `{` (opening curly brace)
    try tokenize_t("'>'", &.{Token.init(.str, 1, 1)}); // single quoting `>` (greater than)
    try tokenize_t("'<'", &.{Token.init(.str, 1, 1)}); // single quoting `<` (less than)
    try tokenize_t("'|'", &.{Token.init(.str, 1, 1)}); // single quoting `|` (pipe symbol)
    try tokenize_t("'}'", &.{Token.init(.str, 1, 1)}); // single quoting `}` (closing curly brace)
    try tokenize_t("' '", &.{Token.init(.str, 1, 1)}); // single quoting ` ` (space)
    try tokenize_t("'\"'", &.{Token.init(.str, 1, 1)}); // single quoting `"` (double quote)
    try tokenize_t("'\\'", &.{Token.init(.str, 1, 1)}); // single quoting `\` (backslash)
    try tokenize_t("'\t'", &.{Token.init(.str, 1, 1)}); // single quoting `\t` (tab)
    try tokenize_t("'\n'", &.{Token.init(.str, 1, 1)}); // single quoting `\n` (newline)
    try tokenize_t("''", &.{Token.init(.str, 1, 0)}); // single quoting `` (empty string)
    try tokenize_t("'A'", &.{Token.init(.str, 1, 1)}); // single quoting `A` (letter A)
}

test "tokenize: double quoting" {
    try tokenize_t("\"!\"", &.{Token.init(.str, 1, 1)}); // double quoting `!` (exclamation mark)
    try tokenize_t("\"$\"", &.{Token.init(.str, 1, 1)}); // double quoting `$` (dollar sign)
    try tokenize_t("\"&\"", &.{Token.init(.str, 1, 1)}); // double quoting `&` (ampersand)
    try tokenize_t("\"(\"", &.{Token.init(.str, 1, 1)}); // double quoting `(` (opening bracket)
    try tokenize_t("\")\"", &.{Token.init(.str, 1, 1)}); // double quoting `)` (closing bracket)
    try tokenize_t("\"*\"", &.{Token.init(.str, 1, 1)}); // double quoting `*` (asterisk)
    try tokenize_t("\";\"", &.{Token.init(.str, 1, 1)}); // double quoting `;` (semicolon)
    try tokenize_t("\"?\"", &.{Token.init(.str, 1, 1)}); // double quoting `?` (question mark)
    try tokenize_t("\"`\"", &.{Token.init(.str, 1, 1)}); // double quoting ``` (tick mark)
    try tokenize_t("\"{\"", &.{Token.init(.str, 1, 1)}); // double quoting `{` (opening curly brace)
    try tokenize_t("\">\"", &.{Token.init(.str, 1, 1)}); // double quoting `>` (greater than)
    try tokenize_t("\"<\"", &.{Token.init(.str, 1, 1)}); // double quoting `<` (less than)
    try tokenize_t("\"|\"", &.{Token.init(.str, 1, 1)}); // double quoting `|` (pipe symbol)
    try tokenize_t("\"}\"", &.{Token.init(.str, 1, 1)}); // double quoting `}` (closing curly brace)
    try tokenize_t("\" \"", &.{Token.init(.str, 1, 1)}); // double quoting ` ` (space)
    try tokenize_t("\"'\"", &.{Token.init(.str, 1, 1)}); // double quoting `'` (single quote)
    try tokenize_t("\"\\\"\"", &.{Token.init(.str, 2, 1)}); // double quoting `"` (double quote)
    try tokenize_t("\"\\\\\"", &.{Token.init(.str, 2, 1)}); // double quoting `\` (backslash)
    try tokenize_t("\"\\$\"", &.{Token.init(.str, 2, 1)}); // double quoting and escaping `$` (dollar sign)
    try tokenize_t("\"\t\"", &.{Token.init(.str, 1, 1)}); // double quoting `\t` (tab)
    try tokenize_t("\"\n\"", &.{Token.init(.str, 1, 1)}); // double quoting `\n` (newline)
    try tokenize_t("\"\"", &.{Token.init(.str, 1, 0)}); // double quoting `\` (empty string)
    try tokenize_t("\"A\"", &.{Token.init(.str, 1, 1)}); // double quoting `A` (letter A)
}

test "tokenize: redirection" {
    try tokenize_t("1>", &.{
        Token.init(.fd, 0, 1),
        Token.init(.gt, 1, 1),
    });
    try tokenize_t("2>", &.{
        Token.init(.fd, 0, 1),
        Token.init(.gt, 1, 1),
    });
    try tokenize_t(">", &.{
        Token.init(.gt, 0, 1),
    });
    try tokenize_t("<", &.{
        Token.init(.lt, 0, 1),
    });
    try tokenize_t("<file.in", &.{
        Token.init(.lt, 0, 1),
        Token.init(.str, 1, 7),
    });
    try tokenize_t(">file.out", &.{
        Token.init(.gt, 0, 1),
        Token.init(.str, 1, 8),
    });
    try tokenize_t("1>file.out", &.{
        Token.init(.fd, 0, 1),
        Token.init(.gt, 1, 1),
        Token.init(.str, 2, 8),
    });
    try tokenize_t("2>file.err", &.{
        Token.init(.fd, 0, 1),
        Token.init(.gt, 1, 1),
        Token.init(.str, 2, 8),
    });
    try tokenize_t("<file.in 1>file.out 2>file.err", &.{
        Token.init(.lt, 0, 1),
        Token.init(.str, 1, 7),
        Token.init(.gap, 8, 1),
        Token.init(.fd, 9, 1),
        Token.init(.gt, 10, 1),
        Token.init(.str, 11, 8),
        Token.init(.gap, 19, 1),
        Token.init(.fd, 20, 1),
        Token.init(.gt, 21, 1),
        Token.init(.str, 22, 8),
    });
}

test "tokenize: redirection: simple" {
    try tokenize_t("foo <file.in", &.{ // stdin redir
        Token.init(.str, 0, 3),
        Token.init(.gap, 3, 1),
        Token.init(.lt, 4, 1),
        Token.init(.str, 5, 7),
    });
    try tokenize_t("foo >file.out", &.{ // stdout redir
        Token.init(.str, 0, 3),
        Token.init(.gap, 3, 1),
        Token.init(.gt, 4, 1),
        Token.init(.str, 5, 8),
    });
    try tokenize_t("foo >>file.out", &.{ // stdout redir
        Token.init(.str, 0, 3),
        Token.init(.gap, 3, 1),
        Token.init(.gt, 4, 1),
        Token.init(.gt, 5, 1),
        Token.init(.str, 6, 8),
    });
    try tokenize_t("foo 1>file.out", &.{ // stdout redir
        Token.init(.str, 0, 3),
        Token.init(.gap, 3, 1),
        Token.init(.fd, 4, 1),
        Token.init(.gt, 5, 1),
        Token.init(.str, 6, 8),
    });
    try tokenize_t("foo 1>>file.out", &.{ // stdout redir
        Token.init(.str, 0, 3),
        Token.init(.gap, 3, 1),
        Token.init(.fd, 4, 1),
        Token.init(.gt, 5, 1),
        Token.init(.gt, 6, 1),
        Token.init(.str, 7, 8),
    });
    try tokenize_t("foo 2>file.err", &.{ // stderr redir
        Token.init(.str, 0, 3),
        Token.init(.gap, 3, 1),
        Token.init(.fd, 4, 1),
        Token.init(.gt, 5, 1),
        Token.init(.str, 6, 8),
    });
    try tokenize_t("foo 2>>file.err", &.{ // stderr redir
        Token.init(.str, 0, 3),
        Token.init(.gap, 3, 1),
        Token.init(.fd, 4, 1),
        Token.init(.gt, 5, 1),
        Token.init(.gt, 6, 1),
        Token.init(.str, 7, 8),
    });
}

test "tokenize: redirection: complex" {
    try tokenize_t("foo <file.in 1>file.out 2>file.err", &.{ // all redirs (cmd first)
        Token.init(.str, 0, 3), // `foo`
        Token.init(.gap, 3, 1), // ` `
        Token.init(.lt, 4, 1), // `<`
        Token.init(.str, 5, 7), // `file.in`
        Token.init(.gap, 12, 1), // ` `
        Token.init(.fd, 13, 1), // `1`
        Token.init(.gt, 14, 1), // `>`
        Token.init(.str, 15, 8), // `file.out`
        Token.init(.gap, 23, 1), // ` `
        Token.init(.fd, 24, 1), // `2`
        Token.init(.gt, 25, 1), // `>`
        Token.init(.str, 26, 8), // `file.err`
    });
    try tokenize_t("<file.in foo 1>file.out 2>file.err", &.{ // all redirs (cmd after in)
        Token.init(.lt, 0, 1), // `<`
        Token.init(.str, 1, 7), // `file.in`
        Token.init(.gap, 8, 1), // ` `
        Token.init(.str, 9, 3), // `foo`
        Token.init(.gap, 12, 1), // ` `
        Token.init(.fd, 13, 1), // `1`
        Token.init(.gt, 14, 1), // `>`
        Token.init(.str, 15, 8), // `file.out
        Token.init(.gap, 23, 1), // `
        Token.init(.fd, 24, 1), // `2
        Token.init(.gt, 25, 1), // `>`
        Token.init(.str, 26, 8), // `file.err
    });
    try tokenize_t("<file.in 1>file.out foo 2>file.err", &.{ // all redirs (cmd before err)
        Token.init(.lt, 0, 1), // `<`
        Token.init(.str, 1, 7), // `file.in`
        Token.init(.gap, 8, 1), // ` `
        Token.init(.fd, 9, 1), // `1`
        Token.init(.gt, 10, 1), // `>`
        Token.init(.str, 11, 8), // `file.out`
        Token.init(.gap, 19, 1), // ` `
        Token.init(.str, 20, 3), // `foo`
        Token.init(.gap, 23, 1), // ` `
        Token.init(.fd, 24, 1), // `2`
        Token.init(.gt, 25, 1), // `>`
        Token.init(.str, 26, 8), // `file.err`
    });
    try tokenize_t("<file.in 1>file.out 2>file.err foo", &.{ // all redirs (cmd last)
        Token.init(.lt, 0, 1), // `<`
        Token.init(.str, 1, 7), // `file.in`
        Token.init(.gap, 8, 1), // ` `
        Token.init(.fd, 9, 1), // `1`
        Token.init(.gt, 10, 1), // `>`
        Token.init(.str, 11, 8), // `file.out`
        Token.init(.gap, 19, 1), // ` `
        Token.init(.fd, 20, 1), // `2`
        Token.init(.gt, 21, 1), // `>`
        Token.init(.str, 22, 8), // `file.err`
        Token.init(.gap, 30, 1), // ` `
        Token.init(.str, 31, 3), // `foo`
    });
    try tokenize_t("foo <file.in bar 1>file.out baz 2>file.err qux quux", &.{ // all redirs (cmd interspersed)
        Token.init(.str, 0, 3), // `foo`
        Token.init(.gap, 3, 1), // ` `
        Token.init(.lt, 4, 1), // `<`
        Token.init(.str, 5, 7), // `file.in`
        Token.init(.gap, 12, 1), // ` `
        Token.init(.str, 13, 3), // `bar`
        Token.init(.gap, 16, 1), // ` `
        Token.init(.fd, 17, 1), // `1`
        Token.init(.gt, 18, 1), // `>`
        Token.init(.str, 19, 8), // `file.out`
        Token.init(.gap, 27, 1), // ` `
        Token.init(.str, 28, 3), // `baz`
        Token.init(.gap, 31, 1), // ` `
        Token.init(.fd, 32, 1), // `2`
        Token.init(.gt, 33, 1), // `>`
        Token.init(.str, 34, 8), // `file.err`
        Token.init(.gap, 42, 1), // ` `
        Token.init(.str, 43, 3), // `qux`
        Token.init(.gap, 46, 1), // ` `
        Token.init(.str, 47, 4), // `quux`
    });
    try tokenize_t("foo <\"a>b\"", &.{ // weird stdin filename
        Token.init(.str, 0, 3), // `foo`
        Token.init(.gap, 3, 1), // ` `
        Token.init(.lt, 4, 1), // `<`
        Token.init(.str, 6, 3), // `"a>b"`
    });
}

test "tokenize: redirection: wannabe false positives" {
    try tokenize_t("foo \"<file.in\" \">file.out\" \"2>file.err\"", &.{ // args look like redirs but we double-quote
        Token.init(.str, 0, 3), // `foo`
        Token.init(.gap, 3, 1), // ` `
        Token.init(.str, 5, 8), // `"<file.in"`
        Token.init(.gap, 14, 1), // ` `
        Token.init(.str, 16, 9), // `">file.out"`
        Token.init(.gap, 26, 1), // ` `
        Token.init(.str, 28, 10), // `"2>file.err"`
    });
    try tokenize_t("foo '<file.in' '>file.out' '2>file.err'", &.{
        Token.init(.str, 0, 3), // `foo`
        Token.init(.gap, 3, 1), // ` `
        Token.init(.str, 5, 8), // `'<file.in'`
        Token.init(.gap, 14, 1), // ` `
        Token.init(.str, 16, 9), // `'>file.out'`
        Token.init(.gap, 26, 1), // ` `
        Token.init(.str, 28, 10), // `'2>file.err'`
    }); // args look like redirs but we single-quote
    try tokenize_t("foo \\<file.in \\>file.out 2\\>file.err", &.{ // args look like redirs but we escape
        Token.init(.str, 0, 3), // `foo`
        Token.init(.gap, 3, 1), // ` `
        Token.init(.str, 5, 1), // `\<`
        Token.init(.str, 6, 7), // `file.in`
        Token.init(.gap, 13, 1), // ` `
        Token.init(.str, 15, 1), // `\>`
        Token.init(.str, 16, 8), // `file.out`
        Token.init(.gap, 24, 1), // ` `
        Token.init(.fd, 25, 1), // `2`
        Token.init(.str, 27, 1), // `\>`
        Token.init(.str, 28, 8), // `file.err`
    });
    try tokenize_t("foo 1 >file.out", &.{
        Token.init(.str, 0, 3), // `foo`
        Token.init(.gap, 3, 1), // ` `
        Token.init(.fd, 4, 1), // `1`
        Token.init(.gap, 5, 1), // ` `
        Token.init(.gt, 6, 1), // `>`
        Token.init(.str, 7, 8), // `file.out`
    }); // stdout redir with '1' as innocent bystander
    try tokenize_t("foo 12>file.out", &.{ // stdout redir with '12' as innocent bystander
        Token.init(.str, 0, 3), // `foo`
        Token.init(.gap, 3, 1), // ` `
        Token.init(.fd, 4, 1), // `1`
        Token.init(.fd, 5, 1), // `2`
        Token.init(.gt, 6, 1), // `>`
        Token.init(.str, 7, 8), // `file.out`
    });
    try tokenize_t("foo 11>file.out", &.{ // stdout redir with '11' as innocent bystander
        Token.init(.str, 0, 3), // `foo`
        Token.init(.gap, 3, 1), // ` `
        Token.init(.fd, 4, 1), // `1`
        Token.init(.fd, 5, 1), // `1`
        Token.init(.gt, 6, 1), // `>`
        Token.init(.str, 7, 8), // `file.out`
    });
}

test "tokenize: codecrafters" {
    try tokenize_t(
        \\echo "hello'example'\\n'test"
    , &.{
        Token.init(.str, 0, 4), // `echo`
        Token.init(.gap, 4, 1), // ` `
        Token.init(.str, 6, 14), // `"hello'example'`
        Token.init(.str, 21, 7), // `\\n'test"`
    });
}

fn merge_tokens(a: Token, b: Token) !Token {
    if (b.bgn < a.bgn) return error.UnorderedTokens;
    if (b.tag != a.tag) return error.NonMatchingTokens;
    if ((a.bgn + a.len) != b.bgn) return error.NonConsecutiveTokens;
    return Token{
        .tag = a.tag,
        .bgn = a.bgn,
        .len = a.len + b.len,
    };
}

const StatementParser = struct {
    allocator: std.mem.Allocator,
    src: []const u8,
    titr: TokenIterator,

    fn init(src: []const u8, allocator: std.mem.Allocator) StatementParser {
        return StatementParser{
            .src = src,
            .titr = TokenIterator.init(src, allocator),
            .allocator = allocator,
        };
    }

    const Elem = union(enum) {
        arg: []const u8,
        stdin_redir: RedirOp,
        stdout_redir: RedirOp,
        stderr_redir: RedirOp,

        pub fn format(value: @This(), comptime fmt: []const u8, options: std.fmt.FormatOptions, writer: anytype) !void {
            _ = fmt;
            _ = options;
            const val = switch (value) {
                inline else => |e| e,
            };
            try writer.writeByte('.');
            try writer.writeAll(@tagName(value));
            try writer.writeAll(" = ");
            try writer.writeAll(util.qs(val));
        }
    };

    const ElemIterator = struct {
        titr: TokenIterator,
        fd_buff: ?Token = null,
        allocator: std.mem.Allocator,
        fn init(titr: TokenIterator, allocator: std.mem.Allocator) ElemIterator {
            return ElemIterator{
                .allocator = allocator,
                .titr = titr,
            };
        }

        // extract *token* body; returned slice points deep into .titr buff
        fn _extract(self: @This(), token: Token) ![]const u8 {
            return try token.get(self.titr.bitr.buff);
        }

        // Like _pull_more_str() but prefix it with all joined source texts of
        // *tokens* and wrap the result as arg element
        fn _pull_arg(self: *@This(), tokens: []const Token) !Elem {
            var al = std.ArrayList(u8).init(self.allocator);
            for (tokens) |t| try al.appendSlice(try t.get(self.titr.bitr.buff));
            const more_str = try self._pull_more_str();
            defer self.allocator.free(more_str);
            try al.appendSlice(more_str);
            return Elem{ .arg = try al.toOwnedSlice() };
        }

        // Allocate memory using self.allocator and fill it with
        // joined bodies of any remaining streak of .str tokens.
        fn _pull_more_str(self: *@This()) ![]const u8 {
            var al = std.ArrayList(u8).init(self.allocator);
            while (true) {
                const t0 = try self.titr.peek() orelse break;
                if (t0.tag != .str and t0.tag != .fd) break;
                try al.appendSlice(try self._extract(t0));
                _ = try self.titr.next(); // just drop t0
            }
            return try al.toOwnedSlice();
        }

        // remove any streak of consecutive gaps
        fn _drop_gaps(self: *@This()) !void {
            while (true) {
                const t0 = try self.titr.peek() orelse break;
                if (t0.tag != .gap) break;
                _ = try self.titr.next(); // just drop t0
            }
        }

        // Pull all necessary tokens to produce a redirect element, with direction *dir*
        // (.out or .in) and *fd_char*, if specified by a previous token.
        fn _pull_redir(self: *@This(), dir: RedirOp.Dir, fd_char: ?u8) !Elem {
            try self._drop_gaps();
            const t0 = try self.titr.peek() orelse return error.InvalidRedirectionPath; // no path provided
            if (t0.tag == .gap) return error.InternalError; // gaps were not dropped?!
            if (t0.tag == .gt) return error.InvalidSyntax; // gt after gt/lt
            if (t0.tag == .lt) return error.InvalidSyntax; // lt after gt/lt
            const path = try self._pull_more_str();
            if (path.len == 0) return error.InvalidRedirectionPath; // zero-length path
            const redir_op = RedirOp{ .dir = dir, .path = path };
            switch (dir) {
                .in => {
                    if (fd_char == null) return Elem{ .stdin_redir = redir_op };
                    return error.InternalError; // dir=.stdin but fd provided?
                },
                .out, .app => {
                    switch (fd_char orelse '1') {
                        '1' => return Elem{ .stdout_redir = redir_op },
                        '2' => return Elem{ .stderr_redir = redir_op },
                        else => return error.UnsupportedFd, // only 1 and 2 are supported
                    }
                },
            }
        }

        fn next(self: *@This()) !?Elem {
            try self._drop_gaps();
            const t0 = try self.titr.next() orelse return null;
            switch (t0.tag) {
                .gap => return error.InternalError, // gaps were not dropped?
                .fd => {
                    // try to pull out redir
                    const t0_body = try self._extract(t0);
                    if (t0_body.len != 1) return error.InternalError; // invalid .fd length
                    const fd_char = t0_body[0];
                    if (try self.titr.peek2_is(.gt, .gt)) {
                        // upcoming `>>`
                        _ = try self.titr.next();
                        _ = try self.titr.next();
                        return try self._pull_redir(.app, fd_char);
                    } else if (try self.titr.peek1_is(.gt)) {
                        // upcoming `>`
                        _ = try self.titr.next();
                        return try self._pull_redir(.out, fd_char);
                    } else if (try self.titr.peek1_is(.lt)) {
                        // upcoming `<`
                        _ = try self.titr.next();
                        return try self._pull_redir(.in, fd_char);
                    } else if (try self.titr.peek1_is(.gap)) {
                        // upcoming .gap
                        try self._drop_gaps();
                        return try self._pull_arg(&.{t0});
                    } else {
                        // upcoming either:
                        //  *  .fd (to be reinterpreted as arg)
                        //  *  or .str,
                        //  *  or nothing
                        return try self._pull_arg(&.{t0});
                    }
                },
                .lt => return try self._pull_redir(.in, null),
                .gt => {
                    if (try self.titr.peek1_is(.gt)) {
                        // upcoming second `>`
                        _ = try self.titr.next();
                        return try self._pull_redir(.app, null);
                    }
                    return try self._pull_redir(.out, null);
                },
                .str => return try self._pull_arg(&.{t0}),
            }
        }
    };

    fn parse(self: *@This()) !Statement {
        var argv_al = std.ArrayList([]const u8).init(self.allocator);
        defer argv_al.deinit();
        var eitr = ElemIterator.init(self.titr, self.allocator);
        var stdin_redir: ?RedirOp = null;
        var stdout_redir: ?RedirOp = null;
        var stderr_redir: ?RedirOp = null;
        while (try eitr.next()) |elem| {
            switch (elem) {
                .arg => |e| try argv_al.append(e),
                .stdin_redir => |e| stdin_redir = e,
                .stdout_redir => |e| stdout_redir = e,
                .stderr_redir => |e| stderr_redir = e,
            }
        }
        return Statement.init(
            try argv_al.toOwnedSlice(),
            stdin_redir,
            stdout_redir,
            stderr_redir,
        );
    }
};

pub fn parse_stmt(src: []const u8, allocator: std.mem.Allocator) !Statement {
    var sitr = StatementParser.init(src, allocator);
    return try sitr.parse();
}

fn fmt_arg(arg: []const u8) []const u8 {
    return util.qs(arg, null);
}

fn assert_optional_redir_ops(oracle: ?RedirOp, result: ?RedirOp) !void {
    const oracle_path = if (oracle) |op| op.path else null;
    const result_path = if (result) |op| op.path else null;
    try util.assert_optional_strings(oracle_path, result_path);
}

fn parse_stmt_t(buff: []const u8, oracle: Statement) !void {
    const result = try parse_stmt(buff, std.heap.page_allocator);
    // FIXME: we're leaking memory a bit (use std.testing.allocator to see)
    // It's not too important for nice cases as in REPL we reset the memory
    // anyway between loops, but it could be a liability.
    //
    // defer {
    //     for (result) |t| std.testing.allocator.free(t);
    //     std.testing.allocator.free(result);
    // }
    //    try std.testing.expectEqual(oracle.tokens.len, result.tokens.len);
    // try std.testing.expectEqualDeep(oracle, result);
    try util.assert_equal_slices(
        []const u8,
        &fmt_arg,
        oracle.argv,
        result.argv,
    );
    try assert_optional_redir_ops(oracle.stdin_redir, result.stdin_redir);
    try assert_optional_redir_ops(oracle.stdout_redir, result.stdout_redir);
    try assert_optional_redir_ops(oracle.stderr_redir, result.stderr_redir);
}

fn parse_stmt_et(buff: []const u8, oracle: anyerror) !void {
    const result = parse_stmt(buff, std.heap.page_allocator);
    try std.testing.expectEqual(oracle, result);
}

test "parse_stmt: epsilon" {
    try parse_stmt_t("", Statement{ .argv = &.{} });
}

test "parse_stmt: whitespace" {
    try parse_stmt_t(" ", Statement{ .argv = &.{} });
    try parse_stmt_t(" ", Statement{ .argv = &.{} });
    try parse_stmt_t("\t", Statement{ .argv = &.{} });
    try parse_stmt_t("  ", Statement{ .argv = &.{} });
    try parse_stmt_t("\t ", Statement{ .argv = &.{} });
    try parse_stmt_t(" \t", Statement{ .argv = &.{} });
    try parse_stmt_t("\t\t", Statement{ .argv = &.{} });
    try parse_stmt_t(" \t ", Statement{ .argv = &.{} });
    try parse_stmt_t("\t \t", Statement{ .argv = &.{} });
}

test "parse_stmt: single quotes" {
    try parse_stmt_et("'", error.MissingEndSingleQuote);
    try parse_stmt_et("'''", error.MissingEndSingleQuote); // quoting `'` (single quote) - not possible
    try parse_stmt_t("'A Z'", Statement{ .argv = &.{"A Z"} }); // quoting `A Z` (A, space, Z)
    try parse_stmt_t("'A\tZ'", Statement{ .argv = &.{"A\tZ"} }); // quoting 0x09 (A, tab, Z)
    try parse_stmt_t("'A`Z'", Statement{ .argv = &.{"A`Z"} }); // quoting 0x96 (A, backtick, Z)
    try parse_stmt_t("'A\\Z'", Statement{ .argv = &.{"A\\Z"} }); // quoting `A\Z` (A, backslash, Z)
    try parse_stmt_t("'A\"Z'", Statement{ .argv = &.{"A\"Z"} }); // quoting `A"Z` (A, double quote, Z)
    try parse_stmt_t("'AMZ'", Statement{ .argv = &.{"AMZ"} }); // quoting `AMZ` (normal string)
}

test "parse_stmt: backslash escaping (generated)" {
    // TEST GENERATED BY utils/tgen.sh
    try parse_stmt_t("\\!", Statement{ .argv = &.{"!"} }); // backslash escaping `!` (exclamation mark)
    try parse_stmt_t("\\$", Statement{ .argv = &.{"$"} }); // backslash escaping `$` (dollar sign)
    try parse_stmt_t("\\&", Statement{ .argv = &.{"&"} }); // backslash escaping `&` (ampersand)
    try parse_stmt_t("\\(", Statement{ .argv = &.{"("} }); // backslash escaping `(` (opening bracket)
    try parse_stmt_t("\\)", Statement{ .argv = &.{")"} }); // backslash escaping `)` (closing bracket)
    try parse_stmt_t("\\*", Statement{ .argv = &.{"*"} }); // backslash escaping `*` (asterisk)
    try parse_stmt_t("\\;", Statement{ .argv = &.{";"} }); // backslash escaping `;` (semicolon)
    try parse_stmt_t("\\?", Statement{ .argv = &.{"?"} }); // backslash escaping `?` (question mark)
    try parse_stmt_t("\\`", Statement{ .argv = &.{"`"} }); // backslash escaping ``` (tick mark)
    try parse_stmt_t("\\{", Statement{ .argv = &.{"{"} }); // backslash escaping `{` (opening curly brace)
    try parse_stmt_t("\\>", Statement{ .argv = &.{">"} }); // backslash escaping `>` (greater than)
    try parse_stmt_t("\\<", Statement{ .argv = &.{"<"} }); // backslash escaping `<` (less than)
    try parse_stmt_t("\\|", Statement{ .argv = &.{"|"} }); // backslash escaping `|` (pipe symbol)
    try parse_stmt_t("\\}", Statement{ .argv = &.{"}"} }); // backslash escaping `}` (closing curly brace)
    try parse_stmt_t("\\ ", Statement{ .argv = &.{" "} }); // backslash escaping ` ` (space)
    try parse_stmt_t("\\'", Statement{ .argv = &.{"'"} }); // backslash escaping `'` (single quote)
    try parse_stmt_t("\\\"", Statement{ .argv = &.{"\""} }); // backslash escaping `"` (double quote)
    try parse_stmt_t("\\\\", Statement{ .argv = &.{"\\"} }); // backslash escaping `\` (backslash)
    try parse_stmt_t("\\\t", Statement{ .argv = &.{"\t"} }); // backslash escaping `\t` (tab)
    try parse_stmt_t("\\\n", Statement{ .argv = &.{"\n"} }); // backslash escaping `\n` (newline)
    try parse_stmt_t("\\A", Statement{ .argv = &.{"A"} }); // backslash escaping `A` (letter A)
}

test "parse_stmt: single quoting (generated)" {
    // TEST GENERATED BY utils/tgen.sh
    try parse_stmt_t("'!'", Statement{ .argv = &.{"!"} }); // single quoting `!` (exclamation mark)
    try parse_stmt_t("'$'", Statement{ .argv = &.{"$"} }); // single quoting `$` (dollar sign)
    try parse_stmt_t("'&'", Statement{ .argv = &.{"&"} }); // single quoting `&` (ampersand)
    try parse_stmt_t("'('", Statement{ .argv = &.{"("} }); // single quoting `(` (opening bracket)
    try parse_stmt_t("')'", Statement{ .argv = &.{")"} }); // single quoting `)` (closing bracket)
    try parse_stmt_t("'*'", Statement{ .argv = &.{"*"} }); // single quoting `*` (asterisk)
    try parse_stmt_t("';'", Statement{ .argv = &.{";"} }); // single quoting `;` (semicolon)
    try parse_stmt_t("'?'", Statement{ .argv = &.{"?"} }); // single quoting `?` (question mark)
    try parse_stmt_t("'`'", Statement{ .argv = &.{"`"} }); // single quoting ``` (tick mark)
    try parse_stmt_t("'{'", Statement{ .argv = &.{"{"} }); // single quoting `{` (opening curly brace)
    try parse_stmt_t("'>'", Statement{ .argv = &.{">"} }); // single quoting `>` (greater than)
    try parse_stmt_t("'<'", Statement{ .argv = &.{"<"} }); // single quoting `<` (less than)
    try parse_stmt_t("'|'", Statement{ .argv = &.{"|"} }); // single quoting `|` (pipe symbol)
    try parse_stmt_t("'}'", Statement{ .argv = &.{"}"} }); // single quoting `}` (closing curly brace)
    try parse_stmt_t("' '", Statement{ .argv = &.{" "} }); // single quoting ` ` (space)
    try parse_stmt_t("'\"'", Statement{ .argv = &.{"\""} }); // single quoting `"` (double quote)
    try parse_stmt_t("'\\'", Statement{ .argv = &.{"\\"} }); // single quoting `\` (backslash)
    try parse_stmt_t("'\t'", Statement{ .argv = &.{"\t"} }); // single quoting `\t` (tab)
    try parse_stmt_t("'\n'", Statement{ .argv = &.{"\n"} }); // single quoting `\n` (newline)
    try parse_stmt_t("''", Statement{ .argv = &.{""} }); // single quoting `\` (empty string)
    try parse_stmt_t("'A'", Statement{ .argv = &.{"A"} }); // single quoting `A` (letter A)
}

test "parse_stmt: double quoting (generated)" {
    // TEST GENERATED BY utils/tgen.sh
    try parse_stmt_t("\"!\"", Statement{ .argv = &.{"!"} }); // double quoting `!` (exclamation mark)
    try parse_stmt_t("\"$\"", Statement{ .argv = &.{"$"} }); // double quoting `$` (dollar sign)
    try parse_stmt_t("\"&\"", Statement{ .argv = &.{"&"} }); // double quoting `&` (ampersand)
    try parse_stmt_t("\"(\"", Statement{ .argv = &.{"("} }); // double quoting `(` (opening bracket)
    try parse_stmt_t("\")\"", Statement{ .argv = &.{")"} }); // double quoting `)` (closing bracket)
    try parse_stmt_t("\"*\"", Statement{ .argv = &.{"*"} }); // double quoting `*` (asterisk)
    try parse_stmt_t("\";\"", Statement{ .argv = &.{";"} }); // double quoting `;` (semicolon)
    try parse_stmt_t("\"?\"", Statement{ .argv = &.{"?"} }); // double quoting `?` (question mark)
    try parse_stmt_t("\"`\"", Statement{ .argv = &.{"`"} }); // double quoting ``` (tick mark)
    try parse_stmt_t("\"{\"", Statement{ .argv = &.{"{"} }); // double quoting `{` (opening curly brace)
    try parse_stmt_t("\">\"", Statement{ .argv = &.{">"} }); // double quoting `>` (greater than)
    try parse_stmt_t("\"<\"", Statement{ .argv = &.{"<"} }); // double quoting `<` (less than)
    try parse_stmt_t("\"|\"", Statement{ .argv = &.{"|"} }); // double quoting `|` (pipe symbol)
    try parse_stmt_t("\"}\"", Statement{ .argv = &.{"}"} }); // double quoting `}` (closing curly brace)
    try parse_stmt_t("\" \"", Statement{ .argv = &.{" "} }); // double quoting ` ` (space)
    try parse_stmt_t("\"'\"", Statement{ .argv = &.{"'"} }); // double quoting `'` (single quote)
    try parse_stmt_t("\"\\\"\"", Statement{ .argv = &.{"\""} }); // double quoting `"` (double quote)
    try parse_stmt_t("\"\\\\\"", Statement{ .argv = &.{"\\"} }); // double quoting `\` (backslash)
    try parse_stmt_t("\"\t\"", Statement{ .argv = &.{"\t"} }); // double quoting `\t` (tab)
    try parse_stmt_t("\"\n\"", Statement{ .argv = &.{"\n"} }); // double quoting `\n` (newline)
    try parse_stmt_t("\"\"", Statement{ .argv = &.{""} }); // double quoting `\` (empty string)
    try parse_stmt_t("\"A\"", Statement{ .argv = &.{"A"} }); // double quoting `A` (letter A)
}

test "parse_stmt: redirection: errors" {
    try parse_stmt_et("1>", error.InvalidRedirectionPath); // error redirection path
    try parse_stmt_et("2>", error.InvalidRedirectionPath); // error redirection path
    try parse_stmt_et(">", error.InvalidRedirectionPath); // error redirection path
    try parse_stmt_et("<", error.InvalidRedirectionPath); // empty redirection path

    try parse_stmt_et("<file.in", error.RedirectWithoutCommand); // no tokens but stdin redir
    try parse_stmt_et(">file.out", error.RedirectWithoutCommand); // no tokens but stdout redir
    try parse_stmt_et("1>file.out", error.RedirectWithoutCommand); // no tokens but stdout redir
    try parse_stmt_et("2>file.err", error.RedirectWithoutCommand); // no tokens but stderr redir
    try parse_stmt_et("<file.in 1>file.out 2>file.err", error.RedirectWithoutCommand); // no tokens but all redirs
}

test "parse_stmt: redirection: simple" {
    try parse_stmt_t("foo <file.in", Statement{ .argv = &.{"foo"}, .stdin_redir = RedirOp{ .path = "file.in", .dir = .in } }); // stdin redir
    try parse_stmt_t("foo < file.in", Statement{ .argv = &.{"foo"}, .stdin_redir = RedirOp{ .path = "file.in", .dir = .in } }); // stdin redir
    try parse_stmt_t("foo <file.in ", Statement{ .argv = &.{"foo"}, .stdin_redir = RedirOp{ .path = "file.in", .dir = .in } }); // stdin redir
    try parse_stmt_t("foo >file.out", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .out } }); // stdout redir
    try parse_stmt_t("foo > file.out", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .out } }); // stdout redir
    try parse_stmt_t("foo >file.out ", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .out } }); // stdout redir
    try parse_stmt_t("foo >>file.out", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .app } }); // stdout redir
    try parse_stmt_t("foo >> file.out", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .app } }); // stdout redir
    try parse_stmt_t("foo >>file.out ", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .app } }); // stdout redir
    try parse_stmt_t("foo 1>file.out", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .out } }); // stdout redir
    try parse_stmt_t("foo 1> file.out", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .out } }); // stdout redir
    try parse_stmt_t("foo 1>file.out ", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .out } }); // stdout redir
    try parse_stmt_t("foo 1>>file.out", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .app } }); // stdout redir
    try parse_stmt_t("foo 1>> file.out", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .app } }); // stdout redir
    try parse_stmt_t("foo 1>>file.out ", Statement{ .argv = &.{"foo"}, .stdout_redir = RedirOp{ .path = "file.out", .dir = .app } }); // stdout redir
    try parse_stmt_t("foo 2>file.err", Statement{ .argv = &.{"foo"}, .stderr_redir = RedirOp{ .path = "file.err", .dir = .out } }); // stderr redir
    try parse_stmt_t("foo 2> file.err", Statement{ .argv = &.{"foo"}, .stderr_redir = RedirOp{ .path = "file.err", .dir = .out } }); // stderr redir
    try parse_stmt_t("foo 2>file.err ", Statement{ .argv = &.{"foo"}, .stderr_redir = RedirOp{ .path = "file.err", .dir = .out } }); // stderr redir
    try parse_stmt_t("foo 2>>file.err", Statement{ .argv = &.{"foo"}, .stderr_redir = RedirOp{ .path = "file.err", .dir = .app } }); // stderr redir
    try parse_stmt_t("foo 2>> file.err", Statement{ .argv = &.{"foo"}, .stderr_redir = RedirOp{ .path = "file.err", .dir = .app } }); // stderr redir
    try parse_stmt_t("foo 2>>file.err ", Statement{ .argv = &.{"foo"}, .stderr_redir = RedirOp{ .path = "file.err", .dir = .app } }); // stderr redir
}

test "parse_stmt: redirection: complex" {
    try parse_stmt_t("foo <file.in 1>file.out 2>file.err", Statement{
        .argv = &.{"foo"},
        .stdin_redir = RedirOp{ .path = "file.in", .dir = .in },
        .stdout_redir = RedirOp{ .path = "file.out", .dir = .out },
        .stderr_redir = RedirOp{ .path = "file.err", .dir = .out },
    }); // all redirs (cmd first)
    try parse_stmt_t("<file.in foo 1>file.out 2>file.err", Statement{
        .argv = &.{"foo"},
        .stdin_redir = RedirOp{ .path = "file.in", .dir = .in },
        .stdout_redir = RedirOp{ .path = "file.out", .dir = .out },
        .stderr_redir = RedirOp{ .path = "file.err", .dir = .out },
    }); // all redirs (cmd after in)
    try parse_stmt_t("<file.in 1>file.out foo 2>file.err", Statement{
        .argv = &.{"foo"},
        .stdin_redir = RedirOp{ .path = "file.in", .dir = .in },
        .stdout_redir = RedirOp{ .path = "file.out", .dir = .out },
        .stderr_redir = RedirOp{ .path = "file.err", .dir = .out },
    }); // all redirs (cmd before err)
    try parse_stmt_t("<file.in 1>file.out 2>file.err foo", Statement{
        .argv = &.{"foo"},
        .stdin_redir = RedirOp{ .path = "file.in", .dir = .in },
        .stdout_redir = RedirOp{ .path = "file.out", .dir = .out },
        .stderr_redir = RedirOp{ .path = "file.err", .dir = .out },
    }); // all redirs (cmd last)
    try parse_stmt_t("foo <file.in bar 1>file.out baz 2>file.err qux quux", Statement{
        .argv = &.{ "foo", "bar", "baz", "qux", "quux" },
        .stdin_redir = RedirOp{ .path = "file.in", .dir = .in },
        .stdout_redir = RedirOp{ .path = "file.out", .dir = .out },
        .stderr_redir = RedirOp{ .path = "file.err", .dir = .out },
    }); // all redirs (cmd interspersed)
    try parse_stmt_t("foo <\"a>b\"", Statement{
        .argv = &.{"foo"},
        .stdin_redir = RedirOp{ .path = "a>b", .dir = .in },
    }); // weird stdin filename

}

test "parse_stmt: redirection: wannabe false positives" {
    try parse_stmt_t("foo \"<file.in\" \">file.out\" \"2>file.err\"", Statement{
        .argv = &.{ "foo", "<file.in", ">file.out", "2>file.err" },
    }); // args look like redirs but we double-quote
    try parse_stmt_t("foo '<file.in' '>file.out' '2>file.err'", Statement{
        .argv = &.{ "foo", "<file.in", ">file.out", "2>file.err" },
    }); // args look like redirs but we single-quote
    try parse_stmt_t("foo \\<file.in \\>file.out 2\\>file.err", Statement{
        .argv = &.{ "foo", "<file.in", ">file.out", "2>file.err" },
    }); // args look like redirs but we escape
    try parse_stmt_t("foo 1 >file.out", Statement{
        .argv = &.{ "foo", "1" },
        .stdout_redir = RedirOp{ .path = "file.out", .dir = .out },
    }); // stdout redir with '1' as innocent bystander
    try parse_stmt_t("foo 12>file.out", Statement{
        .argv = &.{ "foo", "12" },
        .stdout_redir = RedirOp{ .path = "file.out", .dir = .out },
    }); // stdout redir with '12' as innocent bystander
    try parse_stmt_t("foo 11>file.out", Statement{
        .argv = &.{ "foo", "11" },
        .stdout_redir = RedirOp{ .path = "file.out", .dir = .out },
    }); // stdout redir with '11' as innocent bystander
}

test "parse_stmt: codecrafters" {
    try parse_stmt_t(
        \\echo "hello'example'\\n'test"
    , Statement{ .argv = &.{ "echo", "hello'example'\\n'test" } });
}
