const std = @import("std");

pub fn realpathWithTilde(allocator: std.mem.Allocator, path: []const u8) ![]u8 {
    const replaced = try subst_tilde(allocator, path);
    defer allocator.free(replaced);
    return try std.fs.realpathAlloc(allocator, replaced);
}

/// leverage expectEqualStrings() when possible
pub fn assert_equal_slices(
    comptime T: type,
    comptime fmtfn: *const fn (T) []const u8,
    oracle: []const T,
    result: []const T,
) !void {
    var ew = std.io.getStdErr().writer();
    const max_checks = @max(oracle.len, result.len);
    for (0..max_checks) |idx| {
        if (idx >= result.len) {
            const missing = oracle[idx..];
            for (missing) |item| try ew.print(
                "expected but missing: {s} at index {d}\n",
                .{ fmtfn(item), idx },
            );
            return error.TestFailMissingItems;
        }
        if (idx >= oracle.len) {
            const extra = result[idx..];
            for (extra) |item| try ew.print(
                "unexpected extra item: {s} at index {d}\n",
                .{ fmtfn(item), idx },
            );
            return error.TestFailExtraItems;
        }
        try std.testing.expectEqualStrings(
            fmtfn(oracle[idx]),
            fmtfn(result[idx]),
        );
    }
}

/// leverage expectEqualStrings() when possible
pub fn assert_optional_strings(
    oracle: ?[]const u8,
    result: ?[]const u8,
) !void {
    if (oracle == null and result == null) return;
    if (oracle != null and result == null) {
        try std.io.getStdErr().writer().print("expected \"{s}\", got null\n", .{oracle.?});
        return error.TestFailOptinalString;
    }
    if (oracle == null and result != null) {
        try std.io.getStdErr().writer().print("expected null, got \"{s}\"\n", .{result.?});
        return error.TestFailOptinalString;
    }
    try std.testing.expectEqualStrings(oracle.?, result.?);
}

pub fn subst_tilde(allocator: std.mem.Allocator, path: []const u8) ![]u8 {
    var home: []const u8 = undefined;
    var delim: []const u8 = "/";
    var replaced: []u8 = undefined;
    if (!std.mem.startsWith(u8, path, "~")) {
        replaced = try allocator.dupe(u8, path);
        return replaced;
    }
    home = try std.process.getEnvVarOwned(allocator, "HOME");
    defer allocator.free(home);
    if (std.mem.eql(u8, path, "~")) {
        replaced = try allocator.dupe(u8, home);
        return replaced;
    }
    if (!std.mem.startsWith(u8, path, "~/")) {
        replaced = try allocator.dupe(u8, path);
        return replaced;
    }
    if (std.mem.endsWith(u8, home, "/")) delim = "";
    return try std.fmt.allocPrint(allocator, "{s}{s}{s}", .{ home, delim, path[2..] });
}

const QMAX: usize = 64;

pub fn qb(blob: []const u8, max: ?usize) []const u8 {
    var al = std.ArrayList(u8).init(std.heap.page_allocator);
    qblob(blob, max orelse QMAX, al.writer().any()) catch unreachable;
    return al.toOwnedSlice() catch unreachable;
}

pub fn qbs(blobs: []const []const u8, max: ?usize) []const u8 {
    var al = std.ArrayList(u8).init(std.heap.page_allocator);
    al.writer().writeAll("&.{") catch unreachable;
    for (blobs, 0..) |blob, idx| {
        if (idx > 0) al.writer().writeByte(',') catch unreachable;
        qblob(blob, max orelse QMAX, al.writer().any()) catch unreachable;
    }
    al.writer().writeByte('}') catch unreachable;
    return al.toOwnedSlice() catch unreachable;
}

pub fn qs(str: []const u8, max: ?usize) []const u8 {
    var al = std.ArrayList(u8).init(std.heap.page_allocator);
    qstr(str, max orelse QMAX, al.writer().any()) catch unreachable;
    return al.toOwnedSlice() catch unreachable;
}

pub fn qss(strs: []const []const u8, max: ?usize) []const u8 {
    var al = std.ArrayList(u8).init(std.heap.page_allocator);
    al.writer().writeAll("&.{") catch unreachable;
    for (strs, 0..) |str, idx| {
        if (idx > 0) al.writer().writeByte(',') catch unreachable;
        qstr(str, max orelse QMAX, al.writer().any()) catch unreachable;
    }
    al.writer().writeByte('}') catch unreachable;
    return al.toOwnedSlice() catch unreachable;
}

pub fn qblob(blob: []const u8, max: usize, writer: std.io.AnyWriter) !void {
    const last_idx = @min(blob.len, max);
    try writer.print("[{d}]|", .{blob.len});
    try escape_nonascii(blob[0..last_idx], writer);
    if (blob.len > max) _ = try writer.write("..");
    try writer.writeByte('|');
}

pub fn qstr(str: []const u8, max: usize, writer: std.io.AnyWriter) !void {
    const last_idx = @min(str.len, max);
    try writer.writeByte('"');
    try escape(str[0..last_idx], writer);
    if (str.len > max) try writer.print("..({d} more)", .{str.len - max});
    try writer.writeByte('"');
}

pub fn escape(str: []const u8, writer: std.io.AnyWriter) !void {
    for (str) |c| _ = switch (c) {
        '\n' => try writer.write("\\n"),
        '\r' => try writer.write("\\r"),
        '\t' => try writer.write("\\t"),
        '\\' => try writer.write("\\\\"),
        '\'' => try writer.write("\\'"),
        '"' => try writer.write("\\\""),
        else => try writer.writeByte(c),
    };
}

pub fn escape_nonascii(str: []const u8, writer: std.io.AnyWriter) !void {
    for (str) |c| _ = switch (c) {
        '\n' => try writer.write("\\n"),
        '\r' => try writer.write("\\r"),
        '\t' => try writer.write("\\t"),
        '\\' => try writer.write("\\\\"),
        '\'' => try writer.write("\\'"),
        '"' => try writer.write("\\\""),
        else => if (std.ascii.isASCII(c)) try writer.writeByte(c) else try writer.print("\\x{x}", .{c}),
    };
}
